"""FileDoc class for creating model for textual Paradict data file"""
import os.path
import braq
from paradict.document import Document
from paradict import const


class FileDoc(Document):
    """Create a model for a textual Paradict data file"""
    def __init__(self, path, *, schema=None,
                 type_ref=None, obj_builder=None,
                 spacing=1, encoding_mode=const.DATA_MODE):
        """
        Init

        [parameters]
        - path: path string or a pathlib.Path instance
        - schema: a Python dict that serves as schema to validate the sections
        of the document. It is a dictionary of dictionaries, with root keys
        representing a section header.
        - type_ref: optional TypeRef object
        - obj_builder: function that accepts a paradict.box.Obj container and
        returns a fresh new Python object
        - spacing: number of blank lines to place between two adjacent sections
        - encoding_mode: either 1 or 2, to indicate if Python dicts should be
        encoded with the const.DATA_MODE or const.CONFIG_MODE. By default,
        a document's encoding mode is set to const.DATA_MODE
        """
        super().__init__(schema=schema, type_ref=type_ref,
                         obj_builder=obj_builder, spacing=spacing,
                         encoding_mode=encoding_mode)
        self._sections = None
        self._path = path

    @property
    def path(self):
        return self._path

    def get(self, header, skip_comments=True):
        """
        Decode and return the section whose header is provided

        [parameters]
        - header: the string header of the section
        - skip_comments: boolean to tell whether comments should be ignored or not
        """
        self._ensure_sections()
        return super().get(header, skip_comments=skip_comments)

    def set(self, header, body=None):
        """
        Encode and set a new section.
        Note that this operation will edit the linked file.
        Also note that the update method is suited for
        multiple 'set' operations

        [parameters]
        - body: Python dictionary representing the data to encode
        - header: the string header of the section
        """
        self._ensure_sections()
        super().set(header, body)
        self.save()

    def update(self, *sections):
        """
        Encode and set new sections.
        Note that this operation will edit the linked file.

        [parameters]
        - *sections: A section is a 2-tuple: (header, body)
        """
        self._ensure_sections()
        for header, body in sections:
            super().set(header, body)
        self.save()

    def check(self):
        """
        Return the ordered list (a 'tuple' to be precise)
        of section's headers (strings)
        """
        self._ensure_sections()
        return super().check()

    def render(self, *headers):
        """
        Render the entire document or a specific set of sections, i.e.,
        return a textual Paradict string that may be stored in a file.

        [parameters]
        - *headers: Headers of sections to render.
        Omitting this will render the entire document

        [return]
        Returns a string that contains sections (each made of square-brackets delimited header
        and the associated body)
        """
        self._ensure_sections()
        return super().render(*headers)

    def save_to(self, path):
        """Save the document to a new file.
        Here, path is either a path string
        or a pathlib.Path object"""
        self._ensure_sections()
        return super().save_to(path)

    def remove(self, *headers):
        """remove specific sections from both the document model and the linked file"""
        self._ensure_sections()
        super().remove(*headers)
        self.save()

    def clear(self):
        """clear the document (as well the model as the linked file's contents"""
        self._ensure_sections()
        super().clear()
        self.save()

    def load(self):
        """load the document from the linked file"""
        if not os.path.isfile(self._path):
            self._sections = dict()
            return False
        self._sections = braq.read(self._path, end_of_stream="===")
        return True

    def save(self):
        """Save the document in the linked file. Return a confirmation bool"""
        if self._sections is None:
            return False
        s = list()
        for header, body in self._sections.items():
            s.append((header, body))
        braq.write(*s, dest=self._path, spacing=self._spacing)
        return True

    def _ensure_sections(self):
        if self._sections is None:
            self.load()



import unittest
import tempfile
import pathlib
from textwrap import dedent
from paradict import ConfigFile


class TestConfigFile(unittest.TestCase):

    def setUp(self):
        file = tempfile.NamedTemporaryFile(delete=False)
        file.close()
        self._path = file.name
        self._document = ConfigFile(self._path)

    def tearDown(self):
        pathlib.Path(self._path).unlink()

    def test(self):
        body1 = {"_0": "opening line"}
        body2 = {"name": "alex", "pi": 3.14, "_42": 420}
        self._document.update(("", body1), ("header", body2))
        # 1
        with self.subTest():
            r = self._document.get("")
            expected = body1
            self.assertEqual(expected, r)
        # 2
        with self.subTest():
            r = self._document.get("header")
            expected = body2
            self.assertEqual(expected, r)
        # 3
        with self.subTest():
            r = self._document.render()
            expected = """\
            _0 = "opening line"
            
            [header]
            name = "alex"
            pi = 3.14
            _42 = 420"""
            self.assertEqual(dedent(expected), r)


if __name__ == '__main__':
    unittest.main()



import unittest
import pathlib
import tempfile
from textwrap import dedent
from paradict import Document, box


INIT_TEXT = """\
0: "opening line"

[header]
"name": "alex"
"pi": 3.14
42: 420"""


SCHEMA = {"": {0: "str"},
          "header": {"name": "str",
                     "pi": "float",
                     42: "int"}}


SCHEMA_TEXT = """\
0: "str"

[header]
"name": "str"
"pi": "float"
42: "int"
"""


class TestEmptyDocument(unittest.TestCase):

    def setUp(self):
        self._document = Document()

    def test_get_method(self):
        r = self._document.get("header")
        expected = None
        self.assertEqual(expected, r)

    def test_set_method(self):
        data = {"name": "alex", "pi": 3.14, 42: 420}
        self._document.set("header", data)
        r = self._document.get("header")
        expected = data
        self.assertEqual(expected, r)

    def test_check_method(self):
        r = self._document.check()
        expected = tuple()
        self.assertEqual(expected, r)

    def test_render_method(self):
        r = self._document.render()
        expected = str()
        self.assertEqual(expected, r)

    def test_remove_method(self):
        # 1
        with self.subTest("Test 'remove' method without argument"):
            try:
                self._document.remove()
            except Exception as e:
                self.assertTrue(False)
        # 2
        with self.subTest("Test 'remove' method with argument"):
            try:
                self._document.remove("header")
            except Exception as e:
                self.assertTrue(False)

    def test_clear_method(self):
        try:
            self._document.clear()
        except Exception as e:
            self.assertTrue(False)


class TestDocument(unittest.TestCase):

    def setUp(self):
        self._document = Document(INIT_TEXT)

    def test_get_method(self):
        r = self._document.get("header")
        expected = {"name": "alex", "pi": 3.14, 42: 420}
        self.assertEqual(expected, r)

    def test_set_method(self):
        data = {"name": "alex", "pi": 3.14}
        self._document.set("header", data)
        r = self._document.get("header")
        expected = data
        self.assertEqual(expected, r)

    def test_check_method(self):
        r = self._document.check()
        expected = ("", "header")
        self.assertEqual(expected, r)

    def test_render_method(self):
        # 1
        with self.subTest("Render the entire doc"):
            r = self._document.render()
            expected = INIT_TEXT
            self.assertEqual(expected, r)
        # 2
        with self.subTest("Render the unnamed section"):
            r = self._document.render("")
            expected = '0: "opening line"'
            self.assertEqual(expected, r)
        # 3
        with self.subTest("Render the named section"):
            r = self._document.render("header")
            expected = """\
            [header]
            "name": "alex"
            "pi": 3.14
            42: 420"""
            self.assertEqual(dedent(expected), r)

    def test_remove_method(self):
        # 1
        with self.subTest("Test 'remove' method without argument"):
            self._document.remove()
            r = self._document.check()
            expected = ("", "header")
            self.assertEqual(expected, r)
        # 2
        with self.subTest("Test 'remove' method with argument"):
            self._document.remove("header")
            r = self._document.check()
            expected = ("", )
            self.assertEqual(expected, r)

    def test_clear_method(self):
        self._document.clear()
        r = self._document.check()
        expected = tuple()
        self.assertEqual(expected, r)


class TestDocumentWithComment(unittest.TestCase):

    def setUp(self):
        text = """\
        # this is a comment
        0: 42"""
        self._document = Document(dedent(text))

    def test_get_method_default_behavior(self):
        r = self._document.get("")
        expected = {0: 42}
        self.assertEqual(expected, r)

    def test_get_method_with_comments_on(self):
        r = self._document.get("", skip_comments=False)
        r = tuple(r.values())
        expected = ("this is a comment", 42)
        self.assertEqual(expected, r)

    def test_creating_section_with_comment(self):
        body = {box.CommentID(): box.Comment("This is a comment !"),
                0: 42}
        self._document.set("header", body)
        r = self._document.get("header", skip_comments=False)
        r = tuple(r.values())
        expected = ("This is a comment !", 42)
        self.assertEqual(expected, r)


class TestLoadFromFileOperation(unittest.TestCase):

    def setUp(self):
        file = tempfile.NamedTemporaryFile(delete=False)
        file.write(INIT_TEXT.encode("utf-8"))
        file.close()
        self._path = file.name
        self._document = Document()

    def tearDown(self):
        pathlib.Path(self._path).unlink()

    def test(self):
        self._document.load_from(self._path)
        r = self._document.get("header")
        expected = {"name": "alex", "pi": 3.14, 42: 420}
        self.assertEqual(expected, r)


class TestSaveToFileOperation(unittest.TestCase):

    def setUp(self):
        file = tempfile.NamedTemporaryFile(delete=False)
        file.close()
        self._path = file.name
        self._document = Document(INIT_TEXT)

    def tearDown(self):
        pathlib.Path(self._path).unlink()

    def test(self):
        self._document.save_to(self._path)
        with open(self._path, "r", encoding="utf-8") as file:
            r = file.read()
        expected = INIT_TEXT
        self.assertEqual(expected, r)


class TestDocumentWithSchema(unittest.TestCase):

    def setUp(self):
        self._document = Document(INIT_TEXT, schema=SCHEMA)

    def test_schema_property(self):
        r = self._document.schema
        expected = SCHEMA
        self.assertEqual(expected, r)

    def test_validate_method(self):
        # 1
        with self.subTest("Validate the entire doc"):
            r = self._document.validate()
            expected = True
            self.assertEqual(expected, r)
        # 2
        with self.subTest("Validate the unnamed and the named sections"):
            r = self._document.validate("", "header")
            expected = True
            self.assertEqual(expected, r)


class TestDocumentWithEmptySchemaFile(unittest.TestCase):

    def setUp(self):
        file = tempfile.NamedTemporaryFile(delete=False)
        file.close()
        self._path = file.name
        self._document = Document()
        self._document.load_schema(self._path)

    def tearDown(self):
        pathlib.Path(self._path).unlink()

    def test_schema_property(self):
        r = self._document.schema
        expected = dict()
        self.assertEqual(expected, r)

    def test_validate_method(self):
        # 1
        with self.subTest("Validate the entire doc"):
            r = self._document.validate()
            expected = True
            self.assertEqual(expected, r)
        # 2
        with self.subTest("Validate the named section"):
            r = self._document.validate("header")
            expected = True
            self.assertEqual(expected, r)


class TestDocumentWithSchemaFile(unittest.TestCase):

    def setUp(self):
        file = tempfile.NamedTemporaryFile(delete=False)
        file.write(SCHEMA_TEXT.encode("utf-8"))
        file.close()
        self._path = file.name
        self._document = Document()
        self._document.load_schema(self._path)

    def tearDown(self):
        pathlib.Path(self._path).unlink()

    def test_schema_property(self):
        r = self._document.schema
        expected = SCHEMA
        self.assertEqual(expected, r)

    def test_validate_method(self):
        # 1
        with self.subTest("Validate the entire doc"):
            r = self._document.validate()
            expected = True
            self.assertEqual(expected, r)
        # 2
        with self.subTest("Validate the named and the unnamed sections"):
            r = self._document.validate("", "header")
            expected = True
            self.assertEqual(expected, r)


if __name__ == '__main__':
    unittest.main()



"""Modeling a textual Paradict Document made of sections"""
import braq
from paradict.typeref import TypeRef
from paradict import validator, encode, decode
from paradict import errors, const


class Document:
    """Create a model for a textual Paradict Document made of sections."""
    def __init__(self, init_text="", *, schema=None,
                 type_ref=None, obj_builder=None,
                 spacing=1, encoding_mode=const.DATA_MODE):
        """
        Init

        [parameters]
        - init_text: optional TypeRef object
        - schema: a Python dict that serves as schema to validate the sections
        of the document. It is a dictionary of dictionaries, with root keys
        representing a section header.
        - type_ref: optional TypeRef object
        - obj_builder: function that accepts a paradict.box.Obj container and
        returns a fresh new Python object
        - spacing: number of blank lines to place between two adjacent sections
        - encoding_mode: either 1 or 2, to indicate if Python dicts should be
        encoded with the const.DATA_MODE or const.CONFIG_MODE. By default,
        a document's encoding mode is set to const.DATA_MODE
        """
        self._sections = braq.parse(init_text, end_of_stream="===")
        self._schema = schema
        self._type_ref = type_ref if type_ref else TypeRef()
        self._obj_builder = obj_builder
        self._spacing = spacing
        self._encoding_mode = encoding_mode

    @property
    def schema(self):
        return self._schema

    @schema.setter
    def schema(self, val):
        self._schema = val

    @property
    def spacing(self):
        return self._spacing

    @spacing.setter
    def spacing(self, val):
        self._spacing = val

    @property
    def obj_builder(self):
        return self._obj_builder

    @obj_builder.setter
    def obj_builder(self, val):
        self._obj_builder = val

    @property
    def encoding_mode(self):
        return self._encoding_mode

    @property
    def type_ref(self):
        return self._type_ref

    @type_ref.setter
    def type_ref(self, val):
        self._type_ref = val if val else TypeRef()

    def get(self, header, skip_comments=True):
        """
        Decode and return the section whose header is provided

        [parameters]
        - header: the string header of the section
        - skip_comments: boolean to tell whether comments should be ignored or not
        """
        body = self._sections.get(header)
        if not body:
            return None
        body = decode(body, obj_builder=self._obj_builder,
                      skip_comments=skip_comments,
                      type_ref=self._type_ref)
        return body

    def set(self, header, body=None):
        """
        Encode and set a new section

        [parameters]
        - body: Python dictionary representing the data to encode
        - header: the string header of the section
        """
        body = self._type_ref.dict_type if body is None else body
        if type(body) not in self._type_ref.dict_types:
            msg = "The body must be a dictionary"
            raise errors.Error(msg)
        body = encode(body, mode=self._encoding_mode,
                      type_ref=self._type_ref,
                      skip_comments=False,
                      skip_bin_data=False)
        self._sections[header] = body

    def check(self):
        """
        Return the ordered list (a 'tuple' to be precise)
        of section's headers (strings)
        """
        return tuple(self._sections.keys())

    def render(self, *headers):
        """
        Render the entire document or a specific set of sections, i.e.,
        return a textual Paradict string that may be stored in a file.

        [parameters]
        - *headers: Headers of sections to render.
        Omitting this will render the entire document

        [return]
        Returns a string that contains sections (each made of square-brackets delimited header
        and the associated body)
        """
        sections = list()
        cache = set()
        headers = headers if headers else self.check()
        for header in headers:
            if header in cache:
                continue
            else:
                cache.add(header)
            if header not in self._sections:
                continue
            sections.append((header, self._sections.get(header)))
        return braq.render(*sections, spacing=self._spacing)

    def load_schema(self, src):
        """
        Load a schema file

        [parameters]
        - src: either a path, a pathlib.Path object, or a file like object
        """
        r = braq.read(src, end_of_stream="===")
        self._schema = dict()
        for header, body in r.items():
            body = decode(body, type_ref=self._type_ref, obj_builder=self._obj_builder,
                          skip_comments=True)
            self._schema[header] = body

    def validate(self, *headers):
        """
        Validate this entire document or only specific section(s)

        [parameters]
        - *headers: headers to validate. If you ignore this parameter, the document will
        be checked against the schema.

        [return]
        Return true if the document is valid. Raise an exception if the schema is missing
        """
        if self._schema is None:
            msg = "Missing schema"
            raise errors.Error(msg)
        if type(self._schema) not in self._type_ref.dict_types:
            msg = "The schema must be a dictionary whose keys represent the section headers"
            raise errors.Error(msg)
        headers = headers if headers else self.check()
        for header in headers:
            if header not in self._schema:
                continue
            if not validator.validate(self.get(header, skip_comments=True),
                                      self._schema.get(header)):
                return False
        return True

    def load_from(self, path):
        """Load the document from a file by providing
        a path string or pathlib.Path object"""
        self._sections = braq.read(path, end_of_stream="===")

    def save_to(self, path):
        """
        Save the contents of this document to a specific file

        [parameters]
        - path: path to filename. Path may be a pathlib.Path instance
        """
        if not path:
            return False
        sections = list()
        for header, body in self._sections.items():
            sections.append((header, body))
        braq.write(*sections, dest=path, spacing=self._spacing)
        return True

    def remove(self, *headers):
        """
        Remove specific section(s) from this document

        [parameters]
        - *headers: the headers of the sections to remove
        """
        for header in headers:
            try:
                del self._sections[header]
            except KeyError as e:
                pass

    def clear(self):
        """Clear the entire document"""
        self._sections = dict()



"""ConfigFile class for creating model for Paradict configuration file"""
from paradict import const
from paradict.filedoc import FileDoc


class ConfigFile(FileDoc):
    """A configfile is based on FileDoc, itself based on Document.
    The difference between FileDoc and Document is that
    FileDoc implies a physical file.
    ConfigFile is a FileDoc but only for configuration file.
    Its encoding_mode is const.CONFIG_MODE"""
    def __init__(self, path, *, schema=None,
                 type_ref=None, obj_builder=None,
                 spacing=1):
        """
        Init

        [parameters]
        - path: path string or a pathlib.Path instance
        - schema: a Python dict that serves as schema to validate the sections
        of the document. It is a dictionary of dictionaries, with root keys
        representing a section header.
        - type_ref: optional TypeRef object
        - obj_builder: function that accepts a paradict.box.Obj container and
        returns a fresh new Python object
        - spacing: number of blank lines to place between two adjacent sections
        """
        super().__init__(path, schema=schema, type_ref=type_ref,
                         obj_builder=obj_builder, spacing=spacing,
                         encoding_mode=const.CONFIG_MODE)



import unittest
import pathlib
import tempfile
from textwrap import dedent
from paradict import FileDoc


INIT_TEXT = """\
0: "opening line"

[header]
"name": "alex"
"pi": 3.14
42: 420"""


class TestFileDoc(unittest.TestCase):

    def setUp(self):
        file = tempfile.NamedTemporaryFile(delete=False)
        file.write(INIT_TEXT.encode("utf-8"))
        file.close()
        self._path = file.name
        self._document = FileDoc(self._path)

    def tearDown(self):
        pathlib.Path(self._path).unlink()

    def test_get_method(self):
        # 1
        with self.subTest("Test unnamed section"):
            r = self._document.get("")
            expected = {0: "opening line"}
            self.assertEqual(expected, r)
        # 2
        with self.subTest("Test named section"):
            r = self._document.get("header")
            expected = {"name": "alex", "pi": 3.14, 42: 420}
            self.assertEqual(expected, r)

    def test_set_method(self):
        data = {"name": "alx", "pi": 3.14}
        self._document.set("header", data)
        r = read_file(self._path)
        expected = """\
        0: "opening line"
        
        [header]
        "name": "alx"
        "pi": 3.14"""
        self.assertEqual(dedent(expected), r)

    def test_update_method(self):
        body1 = {0: 420}
        body2 = {}
        self._document.update(("", body1), ("header", body2))
        r = read_file(self._path)
        expected = """\
        0: 420
        
        [header]"""
        self.assertEqual(dedent(expected), r)

    def test_check_method(self):
        r = self._document.check()
        expected = ("", "header")
        self.assertEqual(expected, r)

    def test_render_method(self):
        r = self._document.render()
        expected = INIT_TEXT
        self.assertEqual(expected, r)

    def test_remove_method(self):
        self._document.remove("")
        r = read_file(self._path)
        expected = """\
        [header]
        "name": "alex"
        "pi": 3.14
        42: 420"""
        self.assertEqual(dedent(expected), r)

    def test_group_remove_method(self):
        self._document.remove("", "header")
        r = read_file(self._path)
        expected = ""
        self.assertEqual(expected, r)

    def test_clear_method(self):
        self._document.clear()
        r = read_file(self._path)
        expected = ""
        self.assertEqual(expected, r)

    def test_save_method(self):
        """This method is implicitly called by set, update, remove, and clear methods"""
        self.assertTrue(True)

    def test_load_method(self):
        """This method is implicitly called by other methods"""
        self.assertTrue(True)


class TestSaveToMethod(unittest.TestCase):

    def setUp(self):
        # file1
        file1 = tempfile.NamedTemporaryFile(delete=False)
        file1.write(INIT_TEXT.encode("utf-8"))
        file1.close()
        self._path1 = file1.name
        self._document = FileDoc(self._path1)
        # file2
        file2 = tempfile.NamedTemporaryFile(delete=False)
        file2.close()
        self._path2 = file2.name

    def tearDown(self):
        pathlib.Path(self._path1).unlink()
        pathlib.Path(self._path2).unlink()

    def test(self):
        self._document.save_to(self._path2)
        r = read_file(self._path2)
        expected = INIT_TEXT
        self.assertEqual(expected, r)


def read_file(path):
    with open(path, "r", encoding="utf-8") as file:
        data = file.read()
    return data


if __name__ == '__main__':
    unittest.main()
